/*
  test L(#)/M(#)-convexity
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "quadratic.h"
#include "laminar.h"
#include "serialfunc.h"

int  is_mconv_gene2(int dim, double func(int,int*), int lower[], int upper[],
                    int sum);

/* L#-convex */
/* fLg1(x) = tx A x */
static double fLg1(int dim, int x[]) {
    double a0[] = {  8, -1, -2, -3 };
    double a1[] = { -1,  4, -1, -2 };
    double a2[] = { -2, -1,  6, -2 };
    double a3[] = { -3, -2, -2,  9 };
    double *a[] = { a0, a1, a2, a3 };

    assert(dim == 4);  /* check dimension */
    return quadratic(dim, x, a);
}

/* L# and L-convex */
/* fLg2(x) = tx A x */
static double fLg2(int dim, int x[]) {
    double a0[] = {  6, -1, -2, -3 };
    double a1[] = { -1,  4, -1, -2 };
    double a2[] = { -2, -1,  5, -2 };
    double a3[] = { -3, -2, -2,  7 };
    double *a[] = { a0, a1, a2, a3 };

    assert(dim == 4);  /* check dimension */
    return quadratic(dim, x, a);
}

/* M#-convex of laminar type */
static double fMg1(int dim, int x[]) {
    static int v[] = {
        2,40,90, 1,2,   -1,/* y=(x1+x2),    f += 2y^2+40y+90 */
        50,-2,0, 1,     -1,/* y=(x1),       f += 50y^2-2y */
        3,20,0,  1,2,3, -1,/* y=(x1+x2+x3), f += 3y^2+20y */
        1,1,0,   0,     -1,/* y=(x0),       f += y^2 + y */
        0,0,0,             /* end mark */
    };

    assert(dim == 4);  /* check dimension */
    return ilaminar2(dim, x, v);
}


/* M#-convex */
static double fMg2(int dim, int x[]) {
    double r = 0;

    assert(dim == 4);  /* check dimension */
    r +=  2 * x[0] * x[0];
    r +=  3 * x[1] * x[1];
    r +=  2 * (x[2]+20) * (x[2]+20);
    r +=  5 * (x[3]-20) * (x[3]-20);
    return r;
}

/* not M#-convex (i.e., -Mg1(x)) */
static double f1(int dim, int x[]) {
  return - fMg1(dim,x);
}

/* neither L#- nor M#-convex (in fact, L# + M#) */
static double f2(int dim, int x[]) {
    return fLg1(dim, x) + fMg1(dim, x);
}

/* M#-convex */
static double fMg3(int dim, int x[]) {
    assert(dim == 3);  /* check dimension */

    return pow(x[0]+x[1], 2)
        +  pow(x[1]+x[2], 2)
        +  pow(x[2]+x[0], 2);
}

/* neither L#- nor M#-convex */
static double f3(int dim, int x[]) {
    assert(dim == 3);  /* check dimension */

    return sqrt(x[0]+x[1])
        +  sqrt(x[1]+x[2])
        +  sqrt(x[2]+x[0]);
}

/* neither L#- nor M#-convex */
static double f4(int dim, int x[]) {
    assert(dim == 3);  /* check dimension */

    return pow(x[0]+x[1], 4)
        +  pow(x[1]+x[2], 4)
        +  pow(x[2]+x[0], 4);
}

/* not L#-convex */
/* f9(x) = -100 (near (2,2)), fLg1(x) (otherwise) */
static double f9(int dim, int x[]) {
    if (0<x[0] && x[0]<5 && 0<x[1] && x[1]<5) return -100;
    return fLg1(dim, x);
}

static void test(int dim, double f(int,int*), char *f_name,
                 int lower[], int upper[], int sum) {
    int ret;

    counter_reset();
    ret = is_lgconv_gene(dim, f, lower, upper);
    counter_print();
    if (ret) printf(" %s(x) is     an L#-convex func.\n", f_name);
    else     printf(" %s(x) is NOT an L#-convex func.\n", f_name);

    counter_reset();
    ret = is_lconv_gene(dim, f, lower, upper);
    counter_print();
    if (ret) printf(" %s(x) is     an L-convex func.\n", f_name);
    else     printf(" %s(x) is NOT an L-convex func.\n", f_name);

    counter_reset();
    ret = is_mgconv_gene(dim, f, lower, upper);
    counter_print();
    if (ret) printf(" %s(x) is     an M#-convex func.\n", f_name);
    else     printf(" %s(x) is NOT an M#-convex func.\n", f_name);

    counter_reset();
    ret = is_mconv_gene(dim, f, lower, upper, sum);
    counter_print();
    if (ret) printf(" %s(x) is     an M-convex func.\n", f_name);
    else     printf(" %s(x) is NOT an M-convex func.\n", f_name);

    counter_reset();
    ret = is_mconv_gene2(dim, f, lower, upper, sum);
    counter_print();
    if (ret) printf(" %s(x) is     an M-convex func(2).\n", f_name);
    else     printf(" %s(x) is NOT an M-convex func(2).\n", f_name);

    printf("\n");
}

#define TEST(dim,f) test(dim,f,#f,lower,upper,0)

void test_convexity2(const int dummy) {
    int *lower = ivector_new(5 /* dim */, -10, -10, -5, -5);
    int *upper = ivector_new(5 /* dim */, +10, +10, +5, +5);

    verbose_level_off(STATUS);
    verbose_level_off(COUNTER);

    upper[2] = upper[3] = -5;
    lower[2] = lower[3] = +5;
    printf("lower = "); ivector_print(4, lower); printf("\n");
    printf("upper = "); ivector_print(4, upper); printf("\n");
    TEST(4, fLg1);
    TEST(4, fLg2);
    TEST(4, fMg1);
    TEST(4, fMg2);

    TEST(4, f1);
    TEST(4, f2);
    TEST(4, f9);

    TEST(3, fMg3);
    TEST(3, f3);
    TEST(3, f4);

    upper[0] = upper[1] = -1;
    printf("lower = "); ivector_print(4, lower); printf("\n");
    printf("upper = "); ivector_print(4, upper); printf("\n");
    TEST(4, fLg1);
    TEST(4, f9);

    upper[0]++;
    printf("lower = "); ivector_print(4, lower); printf("\n");
    printf("upper = "); ivector_print(4, upper); printf("\n");
    TEST(4, f9);

    free(lower);
    free(upper);
}
