/*
  L#/L/M#-convexity test
  via Hessian matrix on all points inside domain
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"
#include "quadratic.h"

enum class_e { Lg, L, Mg };

/* INPUT  func : pointer to the function */
/* INPUT  x    : test point in domain */
/* WORK   a    : working matrix */
static int lgv(int dim, double func(int,int*), int x[], double **a) {
    if (!make_hesse_lg(dim, func, x, a)) return -1;
    return is_lgconv_quadratic(dim, a);
}

static int lv(int dim, double func(int,int*), int x[], double **a) {
    if (!make_hesse_l(dim, func, x, a)) return -1;
    return is_lconv_quadratic(dim, a);
}

static int mgv(int dim, double func(int,int*), int x[], double **a) {
    if (!make_hesse_mg(dim, func, x, a)) return -1;
    return is_mgconv_quadratic(dim, a);
}

/* decide if the function is L#/L/M#-convex or not */
static int is_convex(int dim, double func(int,int*),
                     int lower[], int upper[], enum class_e class) {
    int ret = TRUE, step;
    int *x = ivector_new(dim);
    double **a = fmatrix_new(dim, dim);
int ok=0, ng=0;

    ivector_copy(dim, x, lower);

    step = upper[0]-lower[0];
    while ((step & (step-1)) != 0) step &= step - 1;
step=2;

    for (step--; ret && step>=1; step/=2) {
        /* step should not be pow of 2 */
        do {
#ifdef DEBUG
            printf("x="); ivector_print(dim, x); printf("\n");
#endif /* DEBUG */
            if      (class == Lg) ret = lgv(dim, func, x, a);
            else if (class == L)  ret =  lv(dim, func, x, a);
	    else if (class == Mg) ret = mgv(dim, func, x, a);
            else                  assert(0 != 0);
            if (ret>0) {
                /*printf("Hesse="); fmatrix_print(dim, a); printf("\n");/**/
            }
            if (!ret) {
                printf("NG, x=");    ivector_print(dim, x);
                printf(", step=%d\n", step);
                printf("Hesse="); fmatrix_print(dim, a); printf("\n");
            }
            if (verbose_level & STATUS && !ret) {
                printf("x=");    ivector_print(dim, x);
                printf(", step=%d\n", step);
                printf("Hesse="); fmatrix_print(dim, a); printf("\n");
            }
if (ret>0) ok++; else if (ret==0) ng++;
ret = TRUE;
        } while (ret && next(dim, x, lower, upper, step));
    }

printf("OK=%4d, NG=%4d, \t", ok, ng);
if (ng>0) ret = FALSE;
    free(a);
    free(x);
    return !!ret;
}


/* decide if the function is L#-convex or not */
int is_lgconv_gene(int dim, double func(int,int*), int lower[], int upper[]) {
    return is_convex(dim, func, lower, upper, Lg);
}

/* decide if the function is L-convex or not */
int  is_lconv_gene(int dim, double func(int,int*), int lower[], int upper[]) {
    return is_convex(dim, func, lower, upper, L);
}

/* decide if the function is M#-convex or not */
int is_mgconv_gene(int dim, double func(int,int*), int lower[], int upper[]) {
    return is_convex(dim, func, lower, upper, Mg);
}


/* experimental implementation */
int is_lgconv_gene2(int dim, double func(int,int*), int init[]) {
    int i;
    int *lower = ivector_new(dim);
    int *upper = ivector_new(dim);

    lgconv_minimize_FW(dim, func, init);
    ivector_copy(dim, lower, init);
    ivector_copy(dim, upper, init);
    for (i=0; i<dim; i++) {
        lower[i]--;
        upper[i]++;
    }
    if (!is_lgconv_gene(dim, func, lower, upper)) {
        return FALSE;
    }
    lgconv_minimize_FW(dim, func, init);
    ivector_copy(dim, lower, init);
    ivector_copy(dim, upper, init);
    for (i=0; i<dim; i++) {
        lower[i]--;
        upper[i]++;
    }
    if (!is_lgconv_gene(dim, func, lower, upper)) {
        return FALSE;
    }
    lgconv_minimize_FW(dim, func, init);
    ivector_copy(dim, lower, init);
    ivector_copy(dim, upper, init);
    for (i=0; i<dim; i++) {
        lower[i]--;
        upper[i]++;
    }
    return TRUE;
}
